/*

Copyright (C) University of Oxford, 2005-2011

University of Oxford means the Chancellor, Masters and Scholars of the
University of Oxford, having an administrative office at Wellington
Square, Oxford OX1 2JD, UK.

This file is part of Chaste.

Chaste is free software: you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published
by the Free Software Foundation, either version 2.1 of the License, or
(at your option) any later version.

Chaste is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details. The offer of Chaste under the terms of the
License is subject to the License being interpreted in accordance with
English Law and subject to any action against the University of Oxford
being under the jurisdiction of the English Courts.

You should have received a copy of the GNU Lesser General Public License
along with Chaste. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef CHASTE_CVODE
#ifndef _ABSTRACTCVODECELL_HPP_
#define _ABSTRACTCVODECELL_HPP_

#include <vector>
#include <string>
#include <boost/shared_ptr.hpp>

// This is only needed to prevent compilation errors on PETSc 2.2/Boost 1.33.1 combo
#include "UblasVectorInclude.hpp"

// Chaste headers
#include "OdeSolution.hpp"
#include "AbstractOdeSystemInformation.hpp"
#include "AbstractStimulusFunction.hpp"
#include "AbstractIvpOdeSolver.hpp"
#include "AbstractParameterisedSystem.hpp"
#include "AbstractCardiacCellInterface.hpp"

// CVODE headers
#include <nvector/nvector_serial.h>

/**
 * A cardiac cell that is designed to be simulated using CVODE.
 * It uses CVODE's vector type natively.
 *
 * Functionality is similar to that provided by AbstractCardiacCell and AbstractOdeSystem,
 * but not identical.  It also includes a direct interface to the CVODE solver, via the
 * Solve methods, since the CvodeAdaptor class doesn't work for us.
 *
 * Assumes that it will be solving stiff systems, so uses BDF/Newton.
 *
 * Note that a call to Solve will initialise the CVODE solver, and free its
 * working memory when done.  There is thus a non-trivial overhead involved.
 *
 * \todo #890 Add an option to just initialise once, and assume subsequent Solve
 *   calls are continuing from where we left off.
 *
 * \todo #889 Integrate this better into the main hierarchy?
 */
class AbstractCvodeCell : public AbstractCardiacCellInterface, public AbstractParameterisedSystem<N_Vector>
{
protected:
    /** Relative tolerance for solver. */
    double mRelTol;
    /** Absolute tolerance for solver. */
    double mAbsTol;

    /** CVODE's internal data. */
    void* mpCvodeMem;
    /**
     * The maximum number of steps to be taken by the solver
     * in its attempt to reach the next output time.
     */
    long int mMaxSteps;

    /** The size of the previous timestep. */
    double mLastInternalStepSize;

    /** The maximum timestep to use. */
    double mMaxDt;

    /**
     * Used to include extra debugging information in exception messages, e.g.
     *  EXCEPTION(DumpState("Gating variable out of range"));
     *
     * @param rMessage  explanatory message to include in exception
     * @param Y  current state variable values (defaults to using #mStateVariables)
     */
    std::string DumpState(const std::string& rMessage,
                          N_Vector Y = NULL);

    /**
     * @b Must be called by concrete subclass constructors to initialise the state
     * variables, after setting #mpSystemInfo.
     */
    void Init();

    /**
     * Create a new N_Vector by copying the given vector.
     * Used by GetStateVariables and SetStateVariablesUsingACopyOfThisVector.
     *
     * @param originalVec  the vector to copy
     */
    N_Vector CopyVector(N_Vector originalVec);

public:
    /**
     * Create a new cardiac cell.
     *
     * @note subclasses @b must call Init() in their constructor after setting #mpSystemInfo.
     *
     * @param pSolver  not used for these cells (they're always solved with CVODE); may be empty
     * @param numberOfStateVariables  the size of the ODE system modelling this cell
     * @param voltageIndex  the index of the transmembrane potential within the vector of state variables
     * @param pIntracellularStimulus  the intracellular stimulus current
     */
    AbstractCvodeCell(boost::shared_ptr<AbstractIvpOdeSolver> pSolver,
                      unsigned numberOfStateVariables,
                      unsigned voltageIndex,
                      boost::shared_ptr<AbstractStimulusFunction> pIntracellularStimulus);

    /**
     * Free the state variables, if they have been set.
     */
    virtual ~AbstractCvodeCell();

    /**
     * Get the current value of the transmembrane potential, as given
     * in our state variable vector.
     */
    double GetVoltage();

    /**
     * Set the transmembrane potential
     * @param voltage  new value
     */
    void SetVoltage(double voltage);

    /**
     * Reset the model's state variables to the default initial conditions.
     */
    void ResetToInitialConditions();

    /**
     * Get the initial conditions for the cell.
     *
     * Creates and returns a fresh N_Vector, which must be destroyed
     * by the caller when finished with.
     */
    N_Vector GetInitialConditions();

    /**
     * Assign a vector to be used for this cell's state.
     *
     * The cell takes responsibility for freeing this vector when it
     * is destroyed.  If the cell already has state, it will be freed.
     *
     * @param stateVars  new state variables vector
     */
    void SetStateVariables(N_Vector stateVars);

    /**
     * Assign a vector to be copied for this cell's state.
     *
     * Caller retains responsibility for freeing the vector.
     *
     * @param stateVars  new state variables vector
     */
    void SetStateVariablesUsingACopyOfThisVector(N_Vector stateVars);

    /**
     * RHS evaluation function, to be provided by subclasses.
     *
     * @param t  the time at which to evaluate the RHS
     * @param y  the values of the state variables at time t
     * @param ydot  to be filled in with the derivatives of the state variables
     */
    virtual void EvaluateRhs(realtype t,
                             N_Vector y,
                             N_Vector ydot)=0;

    //
    // Solver methods
    //

    /**
     * Set the maximum timestep to use for simulating this cell.
     *
     * @param maxDt  the maximum timestep
     */
    void SetTimestep(double maxDt);

    /**
     * Simulate this cell's behaviour between the time interval [tStart, tEnd],
     * updating the internal state variable values.
     *
     * The maximum time step to use is given by #mMaxDt, which defaults to
     * HeartConfig::Instance()->GetPrintingTimeStep() if unset.
     *
     * @param tStart  beginning of the time interval to simulate
     * @param tEnd  end of the time interval to simulate
     */
    virtual void SolveAndUpdateState(double tStart, double tEnd);

    /**
     * Simulates this cell's behaviour between the time interval [tStart, tEnd],
     * and return state variable values.
     *
     * The maximum time step to use will be taken as #mMaxDt.  If this is unset
     * it is the same as the sampling interval, which defaults to
     * HeartConfig::Instance()->GetPrintingTimeStep().
     *
     * @param tStart  beginning of the time interval to simulate
     * @param tEnd  end of the time interval to simulate
     * @param tSamp  sampling interval for returned results (defaults to dt)
     */
    OdeSolution Compute(double tStart, double tEnd, double tSamp=0.0);

    /**
     * Simulates this cell's behaviour between the time interval [tStart, tEnd],
     * but does not update the voltage.
     *
     * @param tStart  beginning of the time interval to simulate
     * @param tEnd  end of the time interval to simulate
     */
    void ComputeExceptVoltage(double tStart, double tEnd);

    /**
     * Set whether to clamp the voltage by setting its derivative to zero.
     * @param clamp whether to clamp
     */
    void SetVoltageDerivativeToZero(bool clamp=true);

    /**
     * Simulate the cell, returning a sampling of the state variables.
     *
     * Uses the current values of the state variables at initial conditions.
     * If the state variables have not been set (either by a prior solve, or
     * a call to SetStateVariables) the initial conditions (given by
     * GetInitialConditions) will be used.
     *
     * The final values of the state variables will also be stored in this object.
     *
     * @param tStart  start time of simulation
     * @param tEnd  end time of simulation
     * @param maxDt  maximum time step to be taken by the adaptive solver
     *   (set this appropriately to avoid missing a stimulus)
     * @param tSamp  sampling interval at which to store results
     */
    OdeSolution Solve(realtype tStart,
                      realtype tEnd,
                      realtype maxDt,
                      realtype tSamp);

    /**
     * Simulate the cell, updating its internal state variables.
     *
     * Uses the current values of the state variables at initial conditions.
     * If the state variables have not been set (either by a prior solve, or
     * a call to SetStateVariables) the initial conditions (given by
     * GetInitialConditions) will be used.
     *
     * @param tStart  start time of simulation
     * @param tEnd  end time of simulation
     * @param maxDt  maximum time step to be taken by the adaptive solver
     *   (set this appropriately to avoid missing a stimulus)
     */
    void Solve(realtype tStart,
               realtype tEnd,
               realtype maxDt);

    /**
     * Change the maximum number of steps to be taken by the solver
     * in its attempt to reach the next output time.  Default is 500 (set by CVODE).
     *
     * @param numSteps new maximum
     */
    void SetMaxSteps(long int numSteps);

    /**
     * Get the maximum number of steps to be taken by the solver
     * in its attempt to reach the next output time.
     */
    long int GetMaxSteps();

    /**
     * Set relative and absolute tolerances; both scalars.
     * If no parameters are given, tolerances will be reset to default values.
     *
     * @param relTol  the relative tolerance for the solver (defaults to 1e-5)
     * @param absTol  the absolute tolerance for the solver (defaults to 1e-7)
     */
    void SetTolerances(double relTol=1e-5, double absTol=1e-7);

    /**
     * Get the relative tolerance.
     */
    double GetRelativeTolerance();

    /**
     * Get the absolute tolerance.
     */
    double GetAbsoluteTolerance();

    /**
     * Get the last step size used internally by CVODE in the last Solve call.
     */
    double GetLastStepSize();

private:

    /**
     * Set up the CVODE data structures needed to solve the given system.
     *
     * @param initialConditions  initial conditions
     * @param tStart  start time of simulation
     * @param maxDt  maximum time step to take
     */
    void SetupCvode(N_Vector initialConditions,
                    realtype tStart,
                    realtype maxDt);

    /** Free CVODE memory after a solve. */
    void FreeCvodeMemory();

    /**
     * Report an error from CVODE.
     *
     * @param flag  CVODE error code
     * @param msg  Our description of the error
     */
    void CvodeError(int flag, const char * msg);

    /**
     * Copy an N_Vector into a std::vector<double>
     *
     * @param v  vector to copy
     */
    std::vector<double> MakeStdVec(N_Vector v);
};


#endif // _ABSTRACTCVODECELL_HPP_
#endif // CHASTE_CVODE
